from silence.settings import settings

from colorama import Fore, Style

import logging
import re

###############################################################################
# Filters and modifies Flask's log records in-place
###############################################################################

# Regex to remove ANSI color codes from log lines
RE_ANSI = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')

RE_LOG = re.compile(r'(.*) - - \[(.*)\] "(\w+) (/.*) HTTP.*" (.*) -.*')

class FlaskFilter(logging.Filter):

    def filter(self, record):
        msg = record.msg

        if msg.startswith(" * Running on"):
            record.msg = msg[3:]
            return True

        msg = RE_ANSI.sub('', msg)
        m = RE_LOG.match(msg)
        if m:
            addr, date, verb, route, code = m.groups()
            
            if route.startswith(settings.API_PREFIX):
                api_web = "[API]"
                api_color = Fore.MAGENTA
            else:
                api_web = "[WEB]"
                api_color = Fore.CYAN

            if code[0] in ('2', '3'):
                code_color = Fore.GREEN
            elif code[0] == '4':
                code_color = Fore.YELLOW
            elif code[0] == '5':
                code_color = Fore.RED
            else:
                code_color = Fore.WHITE

            record.msg = f"{date} | {api_color}{Style.BRIGHT}{api_web}{Style.RESET_ALL} " +\
                  f"{verb} {route} from {addr} - {code_color}{Style.BRIGHT}{code}{Style.RESET_ALL}"
        return True


from flask import jsonify

###############################################################################
# Exceptions used throughout the framework.
# They're very simple so we'll just keep them here for convenience.
###############################################################################

# Something blew up in the database
class DatabaseError(Exception):
    pass

# A SQL string may not be correct
class SQLWarning(Warning):
    pass

# Generic warnings for endpoint creation
class EndpointWarning(Warning):
    pass

# Generic errors for endpoint creation
class EndpointError(Exception):
    pass

# Errors for checking session tokens
class TokenError(Exception):
    pass

# Generic HTTP errors
# Grabbed from https://flask.palletsprojects.com/en/1.1.x/patterns/apierrors/
class HTTPError(Exception):
    def __init__(self, status_code, message=None, exc_class=None, payload=None):
        Exception.__init__(self)
        self.message = message
        self.status_code = status_code
        self.exception = exc_class
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message or f"Error {self.status_code}"
        rv['code'] = self.status_code
        if self.exception:
            rv['exception'] = self.exception
        return rv


from flask import Flask, jsonify, send_from_directory
from flask_cors import CORS
from werkzeug.exceptions import HTTPException
import click

from silence.server.endpoint_loader import load_user_endpoints, load_default_endpoints
from silence.server.api_tree import APITree
from silence.settings import settings
from silence.exceptions import HTTPError
from silence.logging.default_logger import logger
from silence.logging.flask_filter import FlaskFilter

from os.path import join
from os import getcwd
import traceback
import logging

###############################################################################
# The server manager is responsible for setting up the Flask webserver,
# configuring it and deploying the endpoints and web app.
###############################################################################

static_folder = join(getcwd(), "docs") if settings.RUN_WEB else None
APP = Flask(__name__, static_folder=static_folder)
cors = CORS(APP, resources={f"{settings.API_PREFIX}*": {"origins": "*"}})
API_TREE = APITree()

def setup():
    # Configures the web server
    APP.secret_key = settings.SECRET_KEY 
    APP.config["SESSION_TYPE"] = "filesystem"

    # Mute Flask's startup messages
    def noop(*args, **kwargs): pass
    click.echo = noop
    click.secho = noop

    # Add our Flask filter to customize Flask logging messages
    logging.getLogger("werkzeug").addFilter(FlaskFilter())

    # Set up the error handle for our custom exception type
    @APP.errorhandler(HTTPError)
    def handle_HTTPError(error):
        response = jsonify(error.to_dict())
        response.status_code = error.status_code
        return response

    # Set up the generic Exception handler for server errors
    @APP.errorhandler(Exception)
    def handle_generic_error(exc):
        # Pass through our own HTTP error exception
        if isinstance(exc, HTTPError):
            return exc

        # Create a similar JSON response for Werkzeug's exceptions
        if isinstance(exc, HTTPException):
            code = exc.code
            res = jsonify({"message": exc.description, "code": code})
            return res, code
        
        # We're facing an uncontrolled server exception
        logger.exception(exc)

        exc_type = type(exc).__name__
        msg = str(exc)
        err = HTTPError(500, msg, exc_type)
        return handle_HTTPError(err)

    # Load the user-provided API endpoints and the default ones
    if settings.RUN_API:
        load_default_endpoints()
        load_user_endpoints()

    # Load the web static files
    if settings.RUN_WEB:
        logger.debug("Setting up web server")
        @APP.route("/")
        def root():
            return APP.send_static_file("index.html")

        @APP.route("/<path:path>")
        def other_path(path):
            return APP.send_static_file(path)

def run():
    APP.run(
        host=settings.LISTEN_ADDRESS,
        port=settings.HTTP_PORT,
        debug=settings.DEBUG_ENABLED,
        threaded=True,
    )



import dload

from os import listdir
from os.path import join
from shutil import move, rmtree
from secrets import token_urlsafe

def download_from_github(project_name, repo_url):
    dload.git_clone(repo_url, project_name + "/")

    # Unpack it (everything is inside the <name>-master folder)
    branch_folder_name = listdir(project_name)[0]
    
    # Move everything inside that folder outside
    branch_folder = join(project_name, branch_folder_name)
    for elem in listdir(branch_folder):
        move(join(branch_folder, elem), join(project_name, elem))

    # Remove the now empty folder
    rmtree(branch_folder)

    # Generate the random string for the Flask secret key
    # and add it to the settings.py file
    secret_key = token_urlsafe(32)

    with open(join(project_name, "settings.py"), "a", encoding="utf-8") as f:
        f.write("\n\n")
        f.write("# A random string that is used for security purposes\n")
        f.write("# (this has been generated automatically upon project creation)\n")
        f.write(f'SECRET_KEY = "{secret_key}"\n')


from silence.server import manager as server_manager
from silence.logging.default_logger import logger
from silence.settings import settings
from silence import __version__


def handle(argv):
    logger.info(f"Silence v{__version__}")
    logger.debug("Current settings:\n" + str(settings))
    server_manager.setup()
    server_manager.run()